/*
 * j_to_w.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Implementation of jis_to_wide().
 *
 */

#include "defs.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis j_to_w.c 1.14 98/06/01 22:28:14";
#endif

static wide_int (* const kuten_to_wide)(enum char_domain table,
					int code, int code2)
#ifdef __JIS_CONFIG_USE_UNICODE
		= __jis_kuten_to_unicode;
#else
		= __jis_kuten_to_euc;
#endif /* __JIS_CONFIG_USE_UNICODE */

CONFIG_INLINE
static int
figure_jis_state(jis_state *state, char const *s, size_t len) {
	unsigned i;

	assert(len > 0);

	if (state->state == __JIS_ROMAN || state->state == __JIS_ASCII) {
		if (*s == '\016') {
			state->half_katakana = 1;
			return 1;
		}
		if (*s == '\017') {
			state->half_katakana = 0;
			return 1;
		}
	}
		
	if (*s != '\033') {
		return 0;
	}

	if (len < 3) {
		errno = EINVAL;
		return -1;
	}

	switch(s[1]) {
	case '&':
		i = 1;
		do {
			if (i == 5) {
				state->state = __JIS_1990;
				return 6;
			}
			i++;
			if (len == i) {
				errno = EINVAL;
				return -1;
			}
		} while(s[i] == "\33&@\33$B"[i]);
		break;

	case '$':
		switch(s[2]) {
		case '@':
			/* state->state = __JIS_OLD */	
			errno = EILSEQ;
			return -1;
		case 'B':
			state->state = __JIS_1990;
			return 3;
		case '(':
			if (len == 3) {
				errno = EINVAL;
				return -1;
			}
			if (s[3] == 'D') {
				state->state = __JIS_0212;
				return 4;
			}
		}
		break;

	case '(':
		switch(s[2]) {
		case 'J':
		case 'H':	/* Bug compatibility */
			state->state = __JIS_ROMAN;
			return 3;
		case 'B':
			state->state = __JIS_ASCII;
			return 3;
		case 'I':
			state->state = __JIS_KATAKANA;
			return 3;
		}
		break;
	}
	return 0;
}

#ifdef __JIS_CONFIG_HAVE_UNICODE

CONFIG_INLINE
static int
utf8_to_unicode(char const *s, size_t len, long *ucode) {
	unsigned c;
	unsigned mask;
	int n = 0;
	int i;
	long wc;

	if (len < 1) {
		errno = EINVAL;
		return -1;
	}

	c = (unsigned char) s[0];
	mask = 0x80;
	while(c & mask) {
		mask >>= 1;
		if (mask == 0x02) {
			errno = EILSEQ;
			return -1;
		}
		n++;
	}

#if 0
	printf("@@@ c = 0x%02x, mask = 0x%02x, n = %d\n", c, mask, n);
#endif

	if (n == 0) {
		*ucode = (long) c;
		return 1;
	} else if (n == 1) {
		errno = EILSEQ;
		return -1;
	}

	if ((size_t) n > len) {
		errno = EINVAL;
		return -1;
	}

	wc = c & (mask - 1);

	for(i = 1; i < n; i++) {
		c = (unsigned char) s[i];
#if 0
		printf("@@@ c = 0x%02x, wc = 0x%08lx\n", c, wc);
#endif
		if ((c & 0xc0) != 0x80) {
			errno = EILSEQ;
			return -1;
		}
		wc = (wc << 6) | (c & 0x3f);
	}

	*ucode = wc;

	return n;
}

#endif /* __JIS_CONFIG_HAVE_UNICODE */

size_t
jis_to_wide(wide_char *ws, char const *s, size_t len, jis_state *state) {
	int ret;
	enum char_domain table;
	int code;
	int code2 = 0;
	jis_state new_state;
	wide_char wc;
	enum jis_locale_ctypes locale = __jis_get_locale(state);
#ifdef __JIS_CONFIG_HAVE_UNICODE
	long ucode;
#endif

	if (s == NULL) {
		s = "";
		len = 1;
		ws = NULL;
	}
		
	if (len < 1) {
		return 0;
	}

	table = ASCII;
	code = (unsigned char) s[0];
	ret = 1;

#ifdef __JIS_CONFIG_JIS_ESCAPES_IN_ALL_LOCALES
	if (locale != JIS_LOCALE_ASCII && locale != JIS_LOCALE_JIS_ROMAN)
#else
	if (locale == JIS_LOCALE_JIS)
#endif
	{
		new_state = *state;
		while(1) {
			int n = figure_jis_state(&new_state, s, len);
			if (n == 0) {
				break;
			}
			if (n == -1) {
				return (size_t) -1;
			}
			assert((unsigned) n <= len);
			ret += n;
			s += n;
			len -= n;
			code = (unsigned char) s[0];
			if (len < 1) {
				*state = new_state;
				return (size_t) -2;
			}
		}
		
		table = JROMAN;

#if 0
		printf("@@@: s = %p, s[0] = %d, code = %d, len = %ld\n",
		       (void *) s, (int) s[0], (int) code, (long) len);
		fflush(stdout);
#endif

		switch(new_state.state) {
		case __JIS_ASCII:
			table = ASCII;
			break;

		case __JIS_ROMAN:
			table = JROMAN;
			break;

		case __JIS_KATAKANA:
			if (code >= 33 && code <= 126) { 
				table = JKATAKANA;
				code += 128;
			} else {
				table = JROMAN;
			}
			break;

		case __JIS_1983:
		        table = J1983;
			break;

		case __JIS_1990:
			table = J1990;
			break;
			
		case __JIS_0212:
			table = J0212;
			break;
		}
		if (new_state.state >= __JIS_FIRST_TWO_BYTE) {
			if (code >= 33 && code <= 126) {
				if (len < 2) {
					errno = EINVAL;
					return (size_t) -1;
				}
				if (s[1] < 33 || s[1] > 126) {
					errno = EILSEQ;
					return (size_t) -1;
				}
				code = code - 32;
				code2 = (unsigned char) s[1] - 32;
				ret++;
			} else {
				table = JROMAN;
			}
		}
			
		if ((table == JROMAN || table == ASCII)
		    && new_state.half_katakana && code >= 161 && code <= 223) {
			table = JKATAKANA;
		}

		*state = new_state;
	}

	if (state->state < __JIS_FIRST_TWO_BYTE) {
		switch(locale) {
		case JIS_LOCALE_ASCII:
			/* Do nothing, use the default above. */
			break;

		case JIS_LOCALE_JIS_ROMAN:
			table = JROMAN;
			break;

		case JIS_LOCALE_JIS:
			/* Do nothing, already handled above. */
			break;

		case JIS_LOCALE_EUC:
#ifdef CONFIG_EUC_ASCII
			table = ASCII; /* ??? */
#else
			table = JROMAN;
#endif
			if (code >= 161 && code <= 254) {
				if (len < 2) {
					errno = EINVAL;
					return (size_t) -1;
				}
				code2 = (unsigned char) s[1];
				if (code2 < 161 || code2 > 254) {
					errno = EILSEQ;
					return (size_t) -1;
				}
				table = J1990;
				code -= 160;
				code2 -= 160;
				ret++;
			} else if (code == 142) {
				if (len < 2) {
					errno = EINVAL;
					return (size_t) -1;
				}	
				code = (unsigned char) s[1];
				if (code < 161 || code > 223) {
					errno = EILSEQ;
					return (size_t) -1;
				}
				table = JKATAKANA;
				ret++;
			} else if (code == 143) {
				if (len < 3) {
					errno = EINVAL;
					return (size_t) -1;
				}
				code = (unsigned char) s[1];
				code2 = (unsigned char) s[2];
				if (code < 161 || code > 254
				    || code2 < 161 || code2 > 254) {
					errno = EILSEQ;
					return (size_t) -1;
				}
				table = J0212;
				code -= 160;
				code2 -= 160;
				ret += 2;
			}
			break;

		case JIS_LOCALE_SHIFT_JIS:
			table = JKATAKANA;
			if ((code >= 129 && code <= 159)
			    || (code >= 224 && code <= 239)) {
				if (len < 2) {
					errno = EINVAL;
					return (size_t) -1;
				}
				if (code >= 224) {
					code -= (223 - 159);
				}
				code = (code - 129) * 2 + 1;
				code2 = (unsigned char) s[1];
				if (code2 >= 64 && code2 <= 126) {
					code2 -= 63;
				} else if (code2 >= 128 && code <= 252) {
					code2 -= 64;
				} else {
					errno = EILSEQ;
					return (size_t) -1;
				}
				if (code2 > 94) {
					code++;
					code2 -= 94;
				}
				table = J1990;
				ret++;
			}
			break;

#ifdef __JIS_CONFIG_HAVE_UNICODE
		case JIS_LOCALE_UTF8:
			ret = utf8_to_unicode(s, len, &ucode);
			if (ret == -1) {
				return (size_t) -1;
			}
			table = UNICODE;
			break;
#endif /* __JIS_CONFIG_HAVE_UNICODE */
		     
		default:
			abort();
		}
	}

#ifdef __JIS_CONFIG_HAVE_UNICODE
	if (table == UNICODE) {
		wide_int r = jis_unicode_to_wide(ucode);
		if (r == WIDE_EOF) {
			errno = EILSEQ;
			return (size_t) -1;
		}
		wc = (wide_char) r;
	} else
#endif /* __JIS_CONFIG_HAVE_UNICODE */
	{
		assert(code >= 0);
		assert(code2 >= 0);
		assert(code2 < 256);

		wc = kuten_to_wide(table, code, code2);
#if 0
		printf("@@@: (%d, %d, %d) -> %04x\n", (int) table, code, code2,
		       (unsigned) (unsigned short) wc);
#endif
		if (wc == (wide_char) -1) {
			errno = EILSEQ;
			return (size_t) -1;
		}
	}

	if (ws != NULL) {
		*ws = wc;
	}

	if (wc == (wide_char) 0) {
		*state = __jis_initial_state;
#if 0
		return 0; /* What NA1 says we should do */
#endif
	}
	return ret;
}

