/*
 * kindex.h
 *
 * By Ross Ridge
 * Public Domain
 *
 * @(#) myjis kindex.h 1.1 97/02/21 01:44:12
 */

#ifndef __JIS_KINDEX_H__
#define __JIS_KINDEX_H__

#ifdef __GNUC__
typedef list<int> __jis_list_int;
#endif

template <class TYPE>
class qcompare {
public:
	static int fn(TYPE const &p1, TYPE const &p2);
};

template <class TYPE>
int 
qcompare<TYPE>::fn(TYPE const &p1, TYPE const &p2) {
	if (p2 > p1) {
		return -1;
	} else if (p1 == p2) {
		return 0;
	}
	return 1;
};

template <class TYPE>
class qvcompare {
public:
	static int
	fn(void const *p1, void const *p2) {
		return qcompare<TYPE>::fn(*((TYPE *) p1), *((TYPE *) p2));
	}
};

template <class TYPE>
class element {
public:
	TYPE key;
	short line;

	element() {}
	element(TYPE const &k, short l): key(k), line(l) {}
};

template <class TYPE, class COMPARE>
class index_compare: public list<element<TYPE>/**/> {
public:
	typedef element<TYPE> elt;

private:
	static int
	compare_fn(void const *v1, void const *v2) {
		return COMPARE::fn(((elt const *) v1)->key,
				   ((elt const *) v2)->key);
	}

public:
	void
	sort() {
		qsort(compare_fn);
	}

	int
	add(TYPE key, int lineno) {
		return list<elt>::add(elt(key, lineno));
	}


	void 
	operator =(void *p) {
		list<elt>::operator =(p);
	}

	int search(TYPE key, list<int> &lst) const {
	elt ekey(key, 0);
	void *match = bsearch((void const *) &ekey, (void const *) array,
			      len, sizeof(elt), compare_fn);

	if (match == NULL) {
		return 0;
	}

	elt const *p = (elt const *) match;
	do {
		if (lst.add(p->line)) {
			return -1;
		}
		p++;
	} while(p < array + len && COMPARE::fn(key, p->key) == 0);

	p = (elt const *) match;
	while(p > array && COMPARE::fn(key, (--p)->key) == 0) {
		if (lst.add(p->line)) {
			return -1;
		}
	}
	return 0;
}
};

template <class TYPE>
class index: public index_compare<TYPE, qcompare<TYPE>/**/> {
public:
	void 
	operator =(void *p) {
		list<elt>::operator =(p);
	}
};

template <class TYPE>
class auto_free_index: public index<TYPE> {
public:
	void 
	operator =(void *p) {
		index<TYPE>::operator =(p);
	}
	
	~auto_free_index();
};

#include "kindex.c"

#endif /* __JIS_KINDEX_H__ */
