/*
 * list.h
 *
 * By Ross Ridge
 * Public Domain
 *
 * Defines the template list.
 *
 * @(#) myjis list.h 1.1 97/02/21 01:44:15
 *
 */
 
#ifndef __LIST_H__
#define __LIST_H__


template <class TYPE>
class list {
protected:
	TYPE *array;
	int len;
	int left;

	void destroy();
	int copy(list const &l);
	int grow(int new_left);

public:
	list(): array((TYPE *) 0xedcb2345), len(0), left(0) {}

	list(list const &l) {
		if (copy(l)) {
			abort();
		}
	}
	
	list &operator =(list const &l);

	void
	operator =(void *p) {
		assert(p == NULL);
		destroy();
		len = 0;
		left = 0;
	}

	int add(TYPE const &ent);

	int length() const { return len; }

	int extend(int size);

	TYPE const &
	operator [](int index) const {
		assert(index >= 0 && index < len);
		return array[index];
	}

	TYPE &
	select(int index) {
		assert(index >= 0 && index < len);
		return array[index];
	}

	void
	qsort(int (*compare_fn)(void const *, void const *)) {
		::qsort((void *) array, len, sizeof(TYPE), compare_fn);
	}

	~list() { destroy(); }
};

template <class TYPE>
class auto_free_list: public list<TYPE> {
public:
	~auto_free_list();
};

#include "list.c"

/* Appends the contents of one list to another. */

template<class T1, class T2> 
list<T1> &
operator +=(list<T1> &l1, list<T2> const &l2) {
	int i;
	int len = l2.length();
	for(i = 0; i < len; i++) {
		l1.add(l2[i]);
	}
	return l1;
}

/* Deletes the contents of one list from another. */

template<class T1, class T2> 
list<T1> &
operator -=(list<T1> &l1, list<T2> const &l2) {
	int i, j;
	int len1 = l1.length();
	int len2 = l2.length();
	
	if (len2 > 0) {
		list<T1> nl;

		for(i = 0; i < len1; i++) {
			T1 elt = l1[i];
			for(j = 0; j < len2; j++) {
				if (elt == (T1) l2[j]) {
					break;
				}
			}
			if (j == len2) {
				nl.add(elt);
			}
		}

		l1 = nl;
	}

	return l1;
}


#endif /* __LIST_H__ */

/* Local variables: */
/* mode: c++ */
/* End: */
