#
# unitbl.sh
#
# By Ross Ridge
# Public Domain
#
sccs_id='@(#) myjis unitbl.sh 1.10 98/06/01 22:54:34'

. ./config.sh

case "$WIDE_CHARACTER_SET:$INCLUDE_UNICODE_TABLES" in
unicode:*)
	;;
*:yes)
	;;
*)
	cat << 'EOF'
/*
 * This file was automatically generated.
 */

int __jis_dummy_unitbl_var;

EOF
	exit
	;;
esac

maybe_uncompress() {
	if test -f "$1"
	then
		cat "$1"
	elif test -f "$1".gz
	then
		gunzip < "$1".gz
	else
		return 1
	fi
}

case "$1" in
-c)
	comment="-v comment=1"
	;;
*)
	comment=""
	;;
esac

fail=no
for i in JIS0201.TXT JIS0208.TXT JIS0212.TXT
do
	if test ! -f $i -a ! -f $i.gz
	then
		echo "The file $i is missing.  Last time I checked it was available on the" >&2
		echo "Internet as:" >&2

		echo "	ftp://ftp.unicode.org/Public/MAPPINGS/EASTASIA/JIS/$i" >&2
		fail=yes
	fi
done

case $fail in
yes)
	exit 1
	;;
esac

{
	t="[	 ]*"
	conv="2020 - d 100 / 64 * sa 100 % la +"

	maybe_uncompress JIS0201.TXT |
		sed -n "s:^0x\(..\)${t}0x\(....\)$t# \(.*\):16 i \2 p 0 p \1 p [\3] P [;] P A i c:p" | dc

	maybe_uncompress JIS0208.TXT |
		sed -n "s:^0x....${t}0x\(....\)${t}0x\(....\)$t# \(.*\):16 i \2 p 1 p \1 $conv p [\3] P [;] P A i c:p" | dc

	maybe_uncompress JIS0212.TXT |
		sed -n "s:^0x\(....\)${t}0x\(....\)$t# \(.*\):16 i \2 p 2 p \1 $conv p [\3] P [;] P A i c:p" | dc

} | $NEW_AWK -v sccs_id="$sccs_id" $comment '

BEGIN {
	RS = ";"
	FS = "\n"
}

{
#	print ":" $1 ":" $2 ":" $3 ":"
	kuten = $2 * 10000 + $3
	if (uni_to_jis[$1] != "" && uni_to_jis[$1] != kuten) {
		printf("ambiguious: unicode 0x%04X -> kuten %05d, %05d\n",
		       $1, kuten, uni_to_jis[$1]) > "/dev/stderr"
	}
	uni_to_jis[$1] = kuten
	if (comment) {
		uni_name[$1] = $4
	}
}

$2 == 0 {
	if (jisrom_to_uni[$3] != "" && jisrom_to_uni[$3] != $1) {
		printf("ambiguious: JIS roman %3d -> unicode 0x%04X, 0x%04X\n",
		       $3, $1, jisrom_to_uni[$1]) > "/dev/stderr"
	}
	jisrom_to_uni[$3] = $1
}

$2 == 1 {
	if (jis208_to_uni[$3] != "" && jis208_to_uni[$3] != $1) {
		printf("ambiguious: kuten 1%04d -> unicode 0x%04X, 0x%04X\n",
		       $3, $1, jis208_to_uni[$1]) > "/dev/stderr"
	}
	jis208_to_uni[$3] = $1
}

$2 == 2 {
	if (jis212_to_uni[$3] != "" && jis212_to_uni[$3] != $1) {
		printf("ambiguious: kuten 2%04d -> unicode 0x%04X, 0x%04X\n",
		       $3, $1, jis212_to_uni[$1]) > "/dev/stderr"
	}
	jis212_to_uni[$3] = $1
}


function flat_tables() {
	print "short const __jis_wide_to_jis[] = {"
	for(i = 0; i <= 65535; i++) {
		if (uni_to_jis[i] == "") {
			print "\t-1,"
		} else if (comment) {
			printf("\t%5d, /* %s */\n",
			       uni_to_jis[i], uni_name[i]);
		} else {
			printf("\t%d,\n", uni_to_jis[i]);
		}
	}
	print "};"
	print "unsigned short const __jis_1990_to_wide[10000] = {"
	for(i = 0; i < 10000; i++) {
		n = jis208_to_uni[i];
		if (n == "") {
			print "\t0xFFFF,"
		} else if (comment) {
			printf("\t0x%04X, /* %s */\n", n, uni_name[n]);
		} else {
			printf("\t%d,\n", n);
		}
	}		
	print "};"
	print ""
	print "unsigned short const __jis_0212_to_wide[10000] = {"
	for(i = 0; i < 10000; i++) {
		n = jis212_to_uni[i];
		if (n == "") {
			print "\t0xFFFF,"
		} else if (comment) {
			printf("\t0x%04X, /* %s */\n", n, uni_name[n]);
		} else {
			printf("\t%d,\n", n);
		}
	}		
	print "};"
}

function compressed_tables() {
	for(i = 0; i <= 255; i++) {
		n = i * 256
		for(j = 0; j <= 255; j++) {
			if (uni_to_jis[n + j] != "") {
				break
			}
		}
		if (j == 256) {
			uni_to_jis_ind[i] = "jis_wide_to_jis_null"
			continue
		}
		uni_to_jis_ind[i] = "jis_wide_to_jis_" i
		print "static short const jis_wide_to_jis_" i "[256] = {"
		for (j = 0; j <= 255; j++) {
			if (uni_to_jis[n + j] == "") {
				print "\t-1,"
			} else if (comment) {
				printf("\t%5d, /* %s */\n",
				       uni_to_jis[n + j], uni_name[n + j]);
			} else {
				printf("\t%d,\n", uni_to_jis[n + j]);
			}
		}
		print "};"
		print ""
	}

	print "static short const jis_wide_to_jis_null[256] = {"
	for(j = 0; j <= 255; j++) {
		print "\t-1,"
	}
	print "};"
	print ""

	print "short const * const __jis_wide_to_jis[256] = {"
	for(i = 0; i < 256; i++) {
		print "\t" uni_to_jis_ind[i] ","
	}
	print "};"
	print ""

	print "static unsigned short const jis_to_wide_null[100] = {"
	for(j = 0; j <= 99; j++) {
		print "\t0xFFFF,"
	}
	print "};"
	print ""

	for(i = 0; i <= 99; i++) {
		b = i * 100;
		for (j = 0; j <= 99; j++) {
			if (jis208_to_uni[b + j] != "") {
				break;
			}
		}
		if (j == 100) {
			jis208_to_uni_ind[i] = "jis_to_wide_null"
			continue
		}
		jis208_to_uni_ind[i] = "jis_1990_to_wide_" i
		print "static unsigned short const jis_1990_to_wide_" i "[100] = {"
		for (j = 0; j <= 99; j++) {
			n = jis208_to_uni[b + j];
			if (n == "") {
				print "\t0xFFFF,"
			} else if (comment) {
				printf("\t0x%04X, /* %s */\n", n, uni_name[n]);
			} else {
				printf("\t%d,\n", n);
			}
		}
		print "};"
		print ""
	}
			
	print "unsigned short const * const __jis_1990_to_wide[100] = {"
	for(i = 0; i < 100; i++) {
		print "\t" jis208_to_uni_ind[i] ","
	}
	print "};"
	print ""
		
	for(i = 0; i <= 99; i++) {
		b = i * 100;
		for (j = 0; j <= 99; j++) {
			if (jis212_to_uni[b + j] != "") {
				break;
			}
		}
		if (j == 100) {
			jis212_to_uni_ind[i] = "jis_to_wide_null"
			continue
		}
		jis212_to_uni_ind[i] = "jis_0212_to_wide_" i
		print "static unsigned short const jis_0212_to_wide_" i "[100] = {"
		for (j = 0; j <= 99; j++) {
			n = jis212_to_uni[b + j];
			if (n == "") {
				print "\t0xFFFF,"
			} else if (comment) {
				printf("\t0x%04X, /* %s */\n", n, uni_name[n]);
			} else {
				printf("\t%d,\n", n);
			}
		}
		print "};"
		print ""
	}
			
	print "unsigned short const * const __jis_0212_to_wide[100] = {"
	for(i = 0; i < 100; i++) {
		print "\t" jis212_to_uni_ind[i] ","
	}
	print "};"
}

END {
	print "/*"
	print " * Generated automatically from:"
	print " *   ftp://ftp.unicode.org/Public/UNIX/MAPPINGS/EASTASIA/JIS/JIS201.TXT"
	print " *   ftp://ftp.unicode.org/Public/UNIX/MAPPINGS/EASTASIA/JIS/JIS208.TXT"
	print " *   ftp://ftp.unicode.org/Public/UNIX/MAPPINGS/EASTASIA/JIS/JIS212.TXT"
	print " */"
	print ""
	print "#include \"defs.h\""
	print ""
	print "#ifdef CONFIG_SCCS_IDS"
	print "static char const sccs_id[]=\"" sccs_id "\";"
	print "#endif"
	print ""
	print "unsigned short const __jis_roman_to_wide[256] = {"
	for(i = 0; i < 256; i++) {
		n = jisrom_to_uni[i];
		if (n == "") {
			print "\t0xFFFF,"
		} else if (comment) {
			printf("\t0x%04X, /* %s */\n", n, uni_name[n]);
		} else {
			printf("\t%d,\n", n);
		}
	}		
	print "};"
	print ""
	compressed_tables()
}'



