/*
 * w_rtok.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Implementation of wide_roman_to_kana().
 *
 */

#include "defs.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] ="@(#) myjis w_rtok.c 1.7 98/03/18 21:58:46";
#endif

#include <ctype.h>

struct key {
	char str[__JIS_R2K_MAX_ROMAN_LEN + 1];
	size_t len;
};

static int no_possible_completion;

static int
r2k_compare(void const *v1, void const *v2) {
	struct key const *key = v1;
	struct jis_r2k_kana const *p = v2;
	size_t i;
	int n;

	for(i = 0; i < key->len; i++) {
		n = (int) (unsigned char) key->str[i] 
		    - (int) (unsigned char) p->roman[i];
		if (n != 0) {
			return n;
		}
	}
	no_possible_completion = 0;
	n = 0 - (unsigned char) p->roman[i];
	return n;
}
		
int
wide_roman_to_kana(struct jis_r2k_kana const **kana, size_t len,
		   wide_char const *ws, int flags) {
	size_t i;
	struct jis_r2k_kana *p;
	int c;
	wide_char wc;
	struct key key;

	if ((flags & JIS_R2K_ALL) == 0) {
		flags |= JIS_R2K_ALL;
	}

	/*
	 * First try find a match for the roman characters in the table.
	 */

	for (i = 0; i < __JIS_R2K_MAX_ROMAN_LEN && ws[i] != 0; i++) {
		wc = ws[i];
#if 0
		c = wide_to_byte(wc);
		if (c == EOF) {
			break;
		}
#else
		/*
		 *  Assume EUC/Unicode and ASCII.
		 */
		if ((unsigned long) wc > 127) {
			break;
		}
		c = (char) wc;
#endif
		key.str[i] = tolower(c);
	}
	if (i == 0) {
		return 0;
	}
	key.str[i] = '\0';
#if 0
	printf("@@@: key = '%s'\n", key.str);
#endif
	key.len = 1;
	do { 
		no_possible_completion = 1;
		p = bsearch((void const *) &key,
			    (void const *) __jis_r2k_table,
			    R2K_TABLE_LEN, sizeof *__jis_r2k_table,
			    r2k_compare);
		if (p != NULL) {
			break;
		}
	} while(++key.len <= i);

	if (p == NULL) {
		/*
		 * Handle the "n" character specially.
		 */
		if (key.str[0] == 'n') {
			if (len > 0) {
				if (key.str[1] == '\'') {
					kana[0] = &__jis_r2k_n_;
				} else {
					kana[0] = &__jis_r2k_n;
				}
			}
			return 1;
		}

		if (key.str[0] == '-') {
			if (key.str[1] == '-') {
				if (len > 0) {
					kana[0] = &__jis_r2k_long;
				}
				return 1;
			}
			return 0;
		}

		if (no_possible_completion) {
			return -1;
		}
		return 0;
	}

	/*
	 *  Scan back to the first entry in the table that matches.
	 */
	while(p > __jis_r2k_table 
	      && strncmp(key.str, p[-1].roman, key.len) == 0) {
		p--;
	}

	/*
	 *  Now return a list of every entry that matches the roman string
	 *  with the right flags (ie. using the desired romanization system.)
	 */
	i = 0;
	while(p < __jis_r2k_table + R2K_TABLE_LEN
	      && strncmp(key.str, p->roman, key.len) == 0) {
		if (strncmp(key.str, p->roman, strlen(p->roman)) == 0
		    && (p->flags & flags) != 0) {
#if 0
			printf("@@@: roman = %s kana = %04lX %04lX %04lX\n",
			       p->roman, (long) p->kana[0], (long) p->kana[1],
			       (long) p->kana[2]);
#endif
			if (i < len) {
				kana[i] = p;
			}
			i++;
		}
		p++;
	}

	return i;
}
